"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.readDirectory = exports.readdir = exports.readEntireDirectory = exports.deleteEntireDirectory = exports.deleteFile = exports.unlink = exports.writeFile = exports.write = exports.readFile = exports.read = exports.isExecutable = exports.executable = exports.isWritable = exports.writable = exports.isReadable = exports.readable = exports.isPresent = exports.exists = exports.access = exports.mkdirp = void 0;
// builtin
const fs_1 = require("fs");
const path_1 = require("path");
const child_process_1 = require("child_process");
const process_1 = require("process");
const nodeVersion = String(process_1.versions.node || '0');
// external
const errlop_1 = __importDefault(require("errlop"));
const version_compare_1 = __importDefault(require("version-compare"));
/** Make the directory structure */
async function mkdirp(path, opts = {}) {
    // if we already exist, nothing to do
    if (await isPresent(path)) {
        return;
    }
    // create according to node.js version
    if ((0, version_compare_1.default)(nodeVersion, '14.14.0') >= 0) {
        return new Promise(function (resolve, reject) {
            (0, fs_1.mkdir)(path, Object.assign({
                recursive: true,
            }, opts), function (err) {
                if (err)
                    return reject(err);
                resolve();
            });
        });
    }
    else {
        const dir = (0, path_1.dirname)(path);
        await mkdirp(dir, opts);
        return new Promise(function (resolve, reject) {
            (0, fs_1.mkdir)(path, Object.assign({}, opts), function (err) {
                if (err)
                    return reject(err);
                resolve();
            });
        });
    }
}
exports.mkdirp = mkdirp;
/** Check access of a file */
function access(path, mode) {
    return new Promise(function (resolve, reject) {
        (0, fs_1.access)(path, mode, function (err) {
            if (err)
                return reject(err);
            resolve();
        });
    });
}
exports.access = access;
/** Check the existence of a file */
function exists(path) {
    var _a;
    return access(path, (_a = fs_1.constants === null || fs_1.constants === void 0 ? void 0 : fs_1.constants.F_OK) !== null && _a !== void 0 ? _a : 0); // node 4 compat
}
exports.exists = exists;
/** Check the existence status of a file */
function isPresent(path) {
    return exists(path)
        .then(() => true)
        .catch(() => false);
}
exports.isPresent = isPresent;
/** Check the readable status of a file */
function readable(path) {
    var _a;
    return access(path, (_a = fs_1.constants === null || fs_1.constants === void 0 ? void 0 : fs_1.constants.R_OK) !== null && _a !== void 0 ? _a : 4); // node 4 compat
}
exports.readable = readable;
/** Check the readable status of a file */
function isReadable(path) {
    return readable(path)
        .then(() => true)
        .catch(() => false);
}
exports.isReadable = isReadable;
/** Check the writable status of a file */
function writable(path) {
    var _a;
    return access(path, (_a = fs_1.constants === null || fs_1.constants === void 0 ? void 0 : fs_1.constants.W_OK) !== null && _a !== void 0 ? _a : 2); // node 4 compat
}
exports.writable = writable;
/** Check the writable status of a file */
function isWritable(path) {
    return writable(path)
        .then(() => true)
        .catch(() => false);
}
exports.isWritable = isWritable;
/** Check the executable status of a file */
function executable(path) {
    var _a;
    return access(path, (_a = fs_1.constants === null || fs_1.constants === void 0 ? void 0 : fs_1.constants.X_OK) !== null && _a !== void 0 ? _a : 1); // node 4 compat
}
exports.executable = executable;
/** Check the executable status of a file */
function isExecutable(path) {
    return executable(path)
        .then(() => true)
        .catch(() => false);
}
exports.isExecutable = isExecutable;
/** Read a file */
function read(path) {
    return new Promise(function (resolve, reject) {
        (0, fs_1.readFile)(path, { encoding: 'utf8' }, function (err, contents) {
            if (err)
                return reject(new errlop_1.default(`failed to read file at: ${path}`, err));
            return resolve(contents);
        });
    });
}
exports.read = read;
/** Read a file safely */
async function readFile(path) {
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        throw new errlop_1.default(`unable to read the non-existent file: ${path}`, err);
    }
    // check readable
    try {
        await readable(path);
    }
    catch (err) {
        throw new errlop_1.default(`no read permission for the existing file: ${path}`, err);
    }
    // attempt read
    try {
        return await read(path);
    }
    catch (err) {
        throw new errlop_1.default(`failed to read the existing and readable file: ${path}`, err);
    }
}
exports.readFile = readFile;
/** Write contents to a file */
function write(path, contents, mode) {
    return new Promise(function (resolve, reject) {
        (0, fs_1.writeFile)(path, contents, { encoding: 'utf8', mode }, function (err) {
            if (err)
                return reject(new errlop_1.default(`failed to write file at: ${path}`, err));
            return resolve();
        });
    });
}
exports.write = write;
/** Write contents to a file safely */
async function writeFile(path, contents, mode) {
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        // attempt write to non-existent file
        try {
            await write(path, contents, mode);
        }
        catch (err) {
            throw new errlop_1.default(`failed to write the non-existent file: ${path}`, err);
        }
    }
    // check writable
    try {
        await writable(path);
    }
    catch (err) {
        throw new errlop_1.default(`no write permission for the existing file: ${path}`, err);
    }
    // attempt write
    try {
        await write(path, contents, mode);
    }
    catch (err) {
        throw new errlop_1.default(`failed to write the existing and writable file: ${path}`, err);
    }
}
exports.writeFile = writeFile;
/** Delete a file */
function unlink(path) {
    return new Promise(function (resolve, reject) {
        (0, fs_1.unlink)(path, function (err) {
            if (err)
                return reject(new errlop_1.default(`failed to delete file at: ${path}`, err));
            return resolve();
        });
    });
}
exports.unlink = unlink;
/** Delete a file safely */
async function deleteFile(path) {
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        // if it doesn't exist, then we don't care
        return;
    }
    // check writable
    try {
        await writable(path);
    }
    catch (err) {
        if (err.code === 'ENOENT') {
            // if it doesn't exist, then we don't care
            // this may not seem necessary, however it is
            // testen would fail on @bevry/json otherwise
            return;
        }
        throw new errlop_1.default(`no write permission to delete the existing file: ${path}`, err);
    }
    // attempt delete
    try {
        await unlink(path);
    }
    catch (err) {
        if (err.code === 'ENOENT') {
            // if it didn't exist, then we don't care
            // this may not seem necessary, however it is
            // testen would fail on @bevry/json otherwise
            return;
        }
        throw new errlop_1.default(`failed to delete the existing and writable file: ${path}`, err);
    }
}
exports.deleteFile = deleteFile;
/** Remove an entire directory (including nested contents) safely. */
async function deleteEntireDirectory(path, opts = {}) {
    // handle relative and absolute paths correctly
    path = (0, path_1.resolve)(path);
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        // if it doesn't exist, then we don't care
        return;
    }
    // check writable
    try {
        await writable(path);
    }
    catch (err) {
        if (err.code === 'ENOENT') {
            // if it doesn't exist, then we don't care
            // this may not seem necessary, however it is
            // testen would fail on @bevry/json otherwise
            return;
        }
        throw new errlop_1.default(`no write permission to delete the existing file: ${path}`, err);
    }
    // attempt delete
    return new Promise(function (resolve, reject) {
        function next(err) {
            if (err && err.code === 'ENOENT')
                return resolve();
            if (err)
                return reject(err);
            return resolve();
        }
        if ((0, version_compare_1.default)(nodeVersion, '14.14.0') >= 0) {
            Promise.resolve().then(() => __importStar(require('fs'))).then(function ({ rm: _rm }) {
                _rm(path, Object.assign({ recursive: true, force: true, maxRetries: 10 }, opts), next);
            });
        }
        else if ((0, version_compare_1.default)(nodeVersion, '12.16.0') >= 0 &&
            (0, version_compare_1.default)(nodeVersion, '16') < 0) {
            Promise.resolve().then(() => __importStar(require('fs'))).then(function ({ rmdir: _rmdir }) {
                _rmdir(path, Object.assign({ recursive: true, maxRetries: 10 }, opts), next);
            });
        }
        else if ((0, version_compare_1.default)(nodeVersion, '12.10.0') >= 0 &&
            (0, version_compare_1.default)(nodeVersion, '12.16.0') < 0) {
            Promise.resolve().then(() => __importStar(require('fs'))).then(function ({ rmdir: _rmdir }) {
                _rmdir(path, Object.assign({ recursive: true, maxBusyTries: 10 }, opts), next);
            });
        }
        else {
            if (path === '' || path === '/')
                return next(new Error('will not delete root directory'));
            (0, child_process_1.exec)(`rm -rf ${JSON.stringify(path)}`, next);
        }
    });
}
exports.deleteEntireDirectory = deleteEntireDirectory;
/** Read an entire directory (including nested contents) safely. */
async function readEntireDirectory(path) {
    // handle relative and absolute paths correctly
    path = (0, path_1.resolve)(path);
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        throw new errlop_1.default(`unable to read the non-existent directory: ${path}`, err);
    }
    // check readable
    try {
        await readable(path);
    }
    catch (err) {
        throw new errlop_1.default(`no read permission for the directory: ${path}`, err);
    }
    // attempt read
    return new Promise(function (resolve, reject) {
        if ((0, version_compare_1.default)(nodeVersion, '18.17.0') >= 0) {
            (0, fs_1.readdir)(path, { encoding: null, recursive: true, withFileTypes: false }, function (err, files) {
                if (err)
                    return reject(err);
                return resolve(files.sort());
            });
        }
        else {
            // find files and dirs, -f doesn't work on ubuntu
            (0, child_process_1.exec)('find .', { cwd: path }, function (err, stdout) {
                if (err)
                    return reject(err);
                return resolve(stdout
                    .split('\n')
                    .map((p) => p.replace(/^[./\\]+/, '')) // trim . and ./
                    .filter(Boolean)
                    .sort());
            });
        }
    });
}
exports.readEntireDirectory = readEntireDirectory;
/** Read a directory */
function readdir(path) {
    return new Promise(function (resolve, reject) {
        (0, fs_1.readdir)(path, function (err, files) {
            if (err)
                return reject(new errlop_1.default(`failed to read file at: ${path}`, err));
            return resolve(files);
        });
    });
}
exports.readdir = readdir;
/** Read a directory safely */
async function readDirectory(path) {
    // check exists
    try {
        await exists(path);
    }
    catch (err) {
        throw new errlop_1.default(`unable to read the non-existent directory: ${path}`, err);
    }
    // check readable
    try {
        await readable(path);
    }
    catch (err) {
        throw new errlop_1.default(`no read permission for the directory: ${path}`, err);
    }
    // attempt read
    try {
        return await readdir(path);
    }
    catch (err) {
        throw new errlop_1.default(`failed to read the existing and readable directory: ${path}`, err);
    }
}
exports.readDirectory = readDirectory;
